@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2015	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -91,6 +91,8 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 									 POOL_SENT_MESSAGE *message);
 static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *number_of_nodes);
 static int extract_ntuples(char *message);
+static POOL_STATUS close_standby_transactions(POOL_CONNECTION *frontend,
+											  POOL_CONNECTION_POOL *backend);
 
 /*
  * Process Query('Q') message
@@ -110,7 +112,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	POOL_STATUS status;
 	char *string;
 	int lock_kind;
-	int specific_error = 0;
 
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
@@ -125,7 +126,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	}
 
 	/* save last query string for logging purpose */
-	strncpy(query_string_buffer, contents, sizeof(query_string_buffer));
+	strlcpy(query_string_buffer, contents, sizeof(query_string_buffer));
 
 	/* show ps status */
 	query_ps_status(contents, backend);
@@ -205,7 +206,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			{
 				pool_query_context_destroy(query_context);
 				pool_set_skip_reading_from_backends();
-				pool_memory = old_context;
 				return POOL_CONTINUE;
 			}
 		}
@@ -213,7 +213,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		/*
 		 * Start query context
 		 */
-		pool_start_query(query_context, contents, node);
+		pool_start_query(query_context, contents, len, node);
 
 		if (PARALLEL_MODE)
 		{
@@ -224,7 +224,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			if (parallel)
 			{
 				pool_query_context_destroy(query_context);
-				pool_memory = old_context;
 				return status;
 			}
 		}
@@ -278,7 +277,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 			pool_query_context_destroy(query_context);
 			pool_set_skip_reading_from_backends();
-			pool_memory = old_context;
 			return POOL_CONTINUE;
 		}
 
@@ -388,7 +386,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	if (!RAW_MODE)
 	{
 		/* check if query is "COMMIT" or "ROLLBACK" */
-		commit = is_commit_query(node);
+		commit = is_commit_or_rollback_query(node);
 
 		/*
 		 * Query is not commit/rollback
@@ -414,17 +412,27 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 				/* rewrite `now()' to timestamp literal */
 				rewrite_query = rewrite_timestamp(backend, query_context->parse_tree, false, msg);
+
+				/*
+				 * If the query is BEGIN READ WRITE in master/slave mode,
+				 * we send BEGIN instead of it to slaves/standbys.
+				 * original_query which is BEGIN READ WRITE is sent to primary.
+				 * rewritten_query which is BEGIN is sent to standbys.
+				 */
+				if (is_start_transaction_query(query_context->parse_tree) &&
+					is_read_write((TransactionStmt *)query_context->parse_tree) &&
+					MASTER_SLAVE)
+				{
+					rewrite_query = pstrdup("BEGIN");
+				}
+
 				if (rewrite_query != NULL)
 				{
 					query_context->rewritten_query = rewrite_query;
-					len = strlen(rewrite_query) + 1;
+					query_context->rewritten_length = strlen(rewrite_query) + 1;
 				}
-
 			}
 
-			if (query_context->rewritten_query != NULL)
-				string = query_context->rewritten_query;
-
 			/*
 			 * Optimization effort: If there's only one session, we do
 			 * not need to wait for the master node's response, and
@@ -433,54 +441,32 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			if (pool_config->num_init_children == 1)
 			{
 				/* Send query to all DB nodes at once */
-				status = pool_send_and_wait(query_context, string, len, 0, 0, "");
+				status = pool_send_and_wait(query_context, 0, 0);
 				/* free_parser(); */
 				return status;
 			}
 
 			/* Send the query to master node */
-			if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
+			if (pool_send_and_wait(query_context, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
-
-			/* Check specific errors */
-			specific_error = check_errors(backend, MASTER_NODE_ID);
-			if (specific_error)
-			{
-				/* log error message */
-				generate_error_message("SimpleQuery: ", specific_error, contents);
-			}
 		}
 
-		if (specific_error)
+		/*
+		 * Send the query to other than master node.
+		 */
+		if (pool_send_and_wait(query_context, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
-			char msg[1024] = POOL_ERROR_QUERY; /* large enough */
-			int msglen = strlen(msg);
-
-			memset(msg + msglen, 0, sizeof(int));
-
-			/* send query to other nodes */
-			if (pool_send_and_wait(query_context, msg, msglen, -1, MASTER_NODE_ID, "") != POOL_CONTINUE)
-				return POOL_END;
-		}
-		else
-		{
-			/*
-			 * Send the query to other than master node.
-			 */
-			if (pool_send_and_wait(query_context, string, len, -1, MASTER_NODE_ID, "") != POOL_CONTINUE)
-			{
-				pool_query_context_destroy(query_context);
-				return POOL_END;
-			}
+			pool_query_context_destroy(query_context);
+			return POOL_END;
 		}
 
 		/* Send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
-			if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
+			if (pool_send_and_wait(query_context, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				pool_query_context_destroy(query_context);
 				return POOL_END;
@@ -490,7 +476,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	}
 	else
 	{
-		if (pool_send_and_wait(query_context, string, len, 1, MASTER_NODE_ID, "") != POOL_CONTINUE)
+		if (pool_send_and_wait(query_context, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
 			pool_query_context_destroy(query_context);
 			return POOL_END;
@@ -554,18 +540,16 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	node = msg->query_context->parse_tree;
 	query = msg->query_context->original_query;
 	pool_debug("Execute: query: %s", query);
-	strncpy(query_string_buffer, query, sizeof(query_string_buffer));
+	strlcpy(query_string_buffer, query, sizeof(query_string_buffer));
 
-	session_context->query_context = query_context;
 	/*
-	 * Calling pool_where_to_send here is dangerous because the node
-	 * parse/bind has been sent could be change by
-	 * pool_where_to_send() and it leads to "portal not found"
-	 * etc. errors.
+	 * Decide where to send query
 	 */
+	session_context->query_context = query_context;
+	pool_where_to_send(query_context, query, node);
 
 	/* check if query is "COMMIT" or "ROLLBACK" */
-	commit = is_commit_query(node);
+	commit = is_commit_or_rollback_query(node);
 
 	if (REPLICATION || PARALLEL_MODE)
 	{
@@ -575,7 +559,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		if (!commit)
 		{
 			/* Send the query to master node */
-			if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				return POOL_END;
 			}
@@ -598,19 +582,19 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			memset(msg + len, 0, sizeof(int));
 
 			/* send query to other nodes */
-			if (pool_send_and_wait(query_context, msg, len, -1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(query_context, "E", len, msg, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 				return POOL_END;
 		}
 		else
 		{
-			if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 				return POOL_END;
 		}
 		
 		/* send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
-			if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				return POOL_END;
 			}
@@ -618,11 +602,11 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	}
 	else
 	{
-		if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+		if (pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
 			return POOL_END;
 		}
-		if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "E") != POOL_CONTINUE)
+		if (pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
 			return POOL_END;
 		}
@@ -708,16 +692,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 		node = (Node *) lfirst(list_head(parse_tree_list));
 
-		/* If replication mode, check to see what kind of insert lock is
-		 * neccessary.
-		 */
-		if (REPLICATION)
-			insert_stmt_with_lock = need_insert_lock(backend, stmt, node);
+		insert_stmt_with_lock = need_insert_lock(backend, stmt, node);
 
 		/*
 		 * Start query context
 		 */
-		pool_start_query(query_context, pstrdup(stmt), node);
+		pool_start_query(query_context, pstrdup(stmt), strlen(stmt) + 1, node);
 
 		msg = pool_create_sent_message('P', len, contents, 0, name, query_context);
 		if (!msg)
@@ -726,6 +706,8 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			return POOL_END;
 		}
 
+		session_context->uncompleted_message = msg;
+
 		/*
 		 * Decide where to send query
 		 */
@@ -750,8 +732,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			if (rewrite_query != NULL)
 			{
 				int alloc_len = len - strlen(stmt) + strlen(rewrite_query);
-				contents = pool_memory_realloc(session_context->memory_context,
-                                                                      msg->contents, alloc_len);
+				contents = palloc(alloc_len);
 				strcpy(contents, name);
 				strcpy(contents + strlen(name) + 1, rewrite_query);
 				memcpy(contents + strlen(name) + strlen(rewrite_query) + 2,
@@ -769,6 +750,19 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				query_context->rewritten_query = rewrite_query;
 			}
 		}
+
+		/*
+		 * If the query is BEGIN READ WRITE in master/slave mode,
+		 * we send BEGIN instead of it to slaves/standbys.
+		 * original_query which is BEGIN READ WRITE is sent to primary.
+		 * rewritten_query which is BEGIN is sent to standbys.
+		 */
+		if (is_start_transaction_query(query_context->parse_tree) &&
+			is_read_write((TransactionStmt *)query_context->parse_tree) &&
+			MASTER_SLAVE)
+		{
+			query_context->rewritten_query = pstrdup("BEGIN");
+		}
 	}
 	pool_memory = old_context;
 
@@ -812,10 +806,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 
 		if (is_strict_query(query_context->parse_tree))
-		{
 			start_internal_transaction(frontend, backend, query_context->parse_tree);
-			allow_close_transaction = 1;
-		}
 
 		if (insert_stmt_with_lock)
 		{
@@ -829,12 +820,6 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 	}
 
-	/*
-	 * need to set uncompleted_message after calling ReadyForQuery(),
-	 * because the function destroys query context of uncompleted_message.
-	 */
-	session_context->uncompleted_message = msg;
-
 	/*
 	 * Cannot call free_parser() here. Since "string" might be allocated in parser context.
 	 * free_parser();
@@ -847,7 +832,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		 * locks.
 		 */
 		pool_debug("Parse: waiting for master completing the query");
-		if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
+		if (pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
 			pool_query_context_destroy(query_context);
 			return POOL_END;
@@ -878,18 +863,27 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 		if (deadlock_detected)
 		{
+			POOL_QUERY_CONTEXT *error_qc;
+
+			error_qc = pool_init_query_context();
+			pool_start_query(error_qc, POOL_ERROR_QUERY, strlen(POOL_ERROR_QUERY) + 1, node);
+			pool_copy_prep_where(query_context->where_to_send, error_qc->where_to_send);
+			
 			pool_log("Parse: received deadlock error message from master node");
-			if (pool_send_and_wait(query_context, POOL_ERROR_QUERY,
-								   strlen(POOL_ERROR_QUERY)+1, -1,
-								   MASTER_NODE_ID, "") != POOL_CONTINUE)
+
+			if (pool_send_and_wait(error_qc, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
+
+			pool_query_context_destroy(error_qc);
+			pool_set_query_in_progress();
+			session_context->query_context = query_context;
 		}
 		else
 		{
-			if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID) != POOL_CONTINUE)
 			{
 				pool_query_context_destroy(query_context);
 				return POOL_END;
@@ -898,7 +892,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	}
 	else
 	{
-		if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "P") != POOL_CONTINUE)
+		if (pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID) != POOL_CONTINUE)
 		{
 			pool_query_context_destroy(query_context);
 			return POOL_END;
@@ -919,12 +913,11 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 {
 	char *pstmt_name;
 	char *portal_name;
-	char *rewrite_msg = NULL;
+	char *rewrite_msg;
 	POOL_SENT_MESSAGE *parse_msg;
 	POOL_SENT_MESSAGE *bind_msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
-	int insert_stmt_with_lock = 0;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -976,57 +969,34 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	}
 
 	session_context->query_context = query_context;
+	pool_where_to_send(query_context, query_context->original_query,
+					   query_context->parse_tree);
 
 	if (pool_config->load_balance_mode && pool_is_writing_transaction())
 	{
-		pool_where_to_send(query_context, query_context->original_query,
-						   query_context->parse_tree);
-
 		if(parse_before_bind(frontend, backend, parse_msg) != POOL_CONTINUE)
 			return POOL_END;
 	}
 
-	/*
-	 * Start a transaction if necessary in replication mode
-	 */
-	if (REPLICATION)
-	{
-		pool_debug("Bind: checking strict query");
-		if (is_strict_query(query_context->parse_tree))
-		{
-			pool_debug("Bind: strict query");
-			start_internal_transaction(frontend, backend, query_context->parse_tree);
-			allow_close_transaction = 1;
-		}
-
-		pool_debug("Bind: checking insert lock");
-		insert_stmt_with_lock = need_insert_lock(backend, query_context->original_query, query_context->parse_tree);
-		if (insert_stmt_with_lock)
-		{
-			pool_debug("Bind: issuing insert lock");
-			/* issue a LOCK command to keep consistency */
-			if (insert_lock(frontend, backend, query_context->original_query, (InsertStmt *)query_context->parse_tree, insert_stmt_with_lock) != POOL_CONTINUE)
-			{
-				pool_query_context_destroy(query_context);
-				return POOL_END;
-			}
-		}
-	}
-
 	pool_debug("Bind: waiting for master completing the query");
-	if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "B")
+	if (pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
+		if (rewrite_msg != NULL)
+			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
 
-	if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "B")
+	if (pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
+		if (rewrite_msg != NULL)
+			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
-    if(rewrite_msg)
-        free(rewrite_msg);
+
+	if (rewrite_msg != NULL)
+		/* free(rewrite_msg) */;
 
 	return POOL_CONTINUE;
 }
@@ -1078,19 +1048,15 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	}
 
 	session_context->query_context = query_context;
+	pool_where_to_send(query_context, query_context->original_query,
+					   query_context->parse_tree);
 
-	/*
-	 * Calling pool_where_to_send here is dangerous because the node
-	 * parse/bind has been sent could be change by
-	 * pool_where_to_send() and it leads to "portal not found"
-	 * etc. errors.
-	 */
 	pool_debug("Describe: waiting for master completing the query");
-	if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "D")
+	if (pool_extended_send_and_wait(query_context, "D", len, contents, 1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 		return POOL_END;
 
-	if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "D")
+	if (pool_extended_send_and_wait(query_context, "D", len, contents, -1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 		return POOL_END;
 
@@ -1119,16 +1085,21 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		msg = pool_get_sent_message('Q', contents+1);
 		if (!msg)
 			msg = pool_get_sent_message('P', contents+1);
-
 		if (!msg)
-			pool_debug("Close: cannot get parse message");
+		{
+			pool_error("Close: cannot get parse message");
+			return POOL_END;
+		}
 	}
 	/* Portal */
 	else if (*contents == 'P')
 	{
 		msg = pool_get_sent_message('B', contents+1);
 		if (!msg)
-			pool_debug("Close: cannot get bind message");
+		{
+			pool_error("Close: cannot get bind message");
+			return POOL_END;
+		}
 	}
 	else
 	{
@@ -1136,23 +1107,6 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		return POOL_END;
 	}
 
-	/*
-	 * As per the postgresql, calling close on non existing portals is not
-	 * an error. So on the same footings we will ignore all such calls and
-	 * return the close complete message to clients with out going to backend
-	 */
-	if (!msg)
-	{
-		int len = htonl(sizeof(len));
-		pool_set_command_success();
-		pool_unset_query_in_progress();
-
-		pool_write(frontend, "3", 1);
-		pool_write_and_flush(frontend, &len, sizeof(len));
-
-		return POOL_CONTINUE;
-	}
-
 	session_context->uncompleted_message = msg;
 	query_context = msg->query_context;
 
@@ -1166,11 +1120,11 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/* pool_where_to_send(query_context, query_context->original_query, query_context->parse_tree); */
 
 	pool_debug("Close: waiting for master completing the query");
-	if (pool_send_and_wait(query_context, contents, len, 1, MASTER_NODE_ID, "C")
+	if (pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 		return POOL_END;
 
-	if (pool_send_and_wait(query_context, contents, len, -1, MASTER_NODE_ID, "C")
+	if (pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 		return POOL_END;
 
@@ -1374,6 +1328,8 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 			TSTATE(backend, i) = kind;
 
+			pool_debug("ReadyForQuery: transaction state:%c", state);
+
 			/*
 			 * The transaction state to be returned to frontend is
 			 * master's.
@@ -1398,71 +1354,56 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		pool_flush(frontend);
 	}
 
-	if (pool_is_query_in_progress() && pool_is_command_success())
+	if (pool_is_query_in_progress())
 	{
 		node = pool_get_parse_tree();
-		query = pool_get_query_string();
 
-		if (node)
+		if (pool_is_command_success())
 		{
-			/*
-			 * If the query was BEGIN/START TRANSACTION, clear the
-			 * history that we had writing command in the transaction
-			 * and forget the transaction isolation level.
-			 */
-			if (is_start_transaction_query(node))
-			{
-				pool_unset_writing_transaction();
-				pool_unset_failed_transaction();
-				pool_unset_transaction_isolation();
-			}
+			query = pool_get_query_string();
 
-			/*
-			 * SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or SET
-			 * SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL
-			 * SERIALIZABLE, remember it.
-			 */
-			else if (is_set_transaction_serializable(node, query))
+			if (node)
 			{
-				pool_set_transaction_isolation(POOL_SERIALIZABLE);
-			}
-
-			/*
-			 * If 2PC commands, automatically close transaction on standbys since
-			 * 2PC commands close transaction on primary.
-			 */
-			else if (is_2pc_transaction_query(node))
-			{
-				for (i=0;i<NUM_BACKENDS;i++)
-				{
-					if (CONNECTION_SLOT(backend, i) &&
-						TSTATE(backend, i) == 'T' &&
-						BACKEND_INFO(i).backend_status == CON_UP &&
-						(MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != i)
-					{
-						per_node_statement_log(backend, i, "COMMIT");
-						if (do_command(frontend, CONNECTION(backend, i), "COMMIT", MAJOR(backend), 
-									   MASTER_CONNECTION(backend)->pid,
-									   MASTER_CONNECTION(backend)->key, 0) != POOL_CONTINUE)
-						{
-							return POOL_END;
-						}
-					}
-				}
-			}
-
-			/*
-			 * If the query was not READ SELECT, remember that we had
-			 * a write query in this transaction.
-			 */
-			else if (!is_select_query(node, query))
-			{
-				/* However, if the query is "SET TRANSACTION READ ONLY" or its variant,
-				 * don't set it.
+				/*
+				 * If the query was BEGIN/START TRANSACTION, clear the
+				 * history that we had a writing command in the transaction
+				 * and forget the transaction isolation level.
 				 */
-				if (!pool_is_transaction_read_only(node))
+				if (is_start_transaction_query(node))
+				{
+					pool_unset_writing_transaction();
+					pool_unset_failed_transaction();
+					pool_unset_transaction_isolation();
+				}
+
+				/*
+				 * SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or SET
+				 * SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL
+				 * SERIALIZABLE, remember it.
+				 */
+				else if (is_set_transaction_serializable(node, query))
+				{
+					pool_set_transaction_isolation(POOL_SERIALIZABLE);
+				}
+
+				/*
+				 * If 2PC commands has been executed, automatically close
+				 * transactions on standbys if there is any open
+				 * transaction since 2PC commands close transaction on
+				 * primary.
+				 */
+				else if (is_2pc_transaction_query(node))
+				{
+					if (close_standby_transactions(frontend, backend) != POOL_CONTINUE)
+						return POOL_END;
+				}
+
+				/*
+				 * If the query was not READ SELECT, remember that we had
+				 * a write query in this transaction.
+				 */
+				else if (!is_select_query(node, query))
 				{
-					pool_debug("not SET TRANSACTION READ ONLY");
 					pool_set_writing_transaction();
 				}
 			}
@@ -1470,10 +1411,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		pool_unset_query_in_progress();
 	}
 
-	if (session_context->query_context && !pool_is_doing_extended_query_message())
+	if (!pool_is_doing_extended_query_message())
 	{
 		if (!(node && IsA(node, PrepareStmt)))
-			pool_query_context_destroy(session_context->query_context);
+			pool_query_context_destroy(pool_get_session_context()->query_context);
 	}
 
 	sp = MASTER_CONNECTION(backend)->sp;
@@ -1488,6 +1429,33 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
+/*
+ * Close running transactions on standbys.
+ */
+static POOL_STATUS close_standby_transactions(POOL_CONNECTION *frontend,
+											  POOL_CONNECTION_POOL *backend)
+{
+	int i;
+
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (CONNECTION_SLOT(backend, i) &&
+			TSTATE(backend, i) == 'T' &&
+			BACKEND_INFO(i).backend_status == CON_UP &&
+			(MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != i)
+		{
+			per_node_statement_log(backend, i, "COMMIT");
+			if (do_command(frontend, CONNECTION(backend, i), "COMMIT", MAJOR(backend), 
+						   MASTER_CONNECTION(backend)->pid,
+						   MASTER_CONNECTION(backend)->key, 0) != POOL_CONTINUE)
+			{
+				return POOL_END;
+			}
+		}
+	}
+	return POOL_CONTINUE;
+}
+
 POOL_STATUS ParseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
 	POOL_SESSION_CONTEXT *session_context;
@@ -1811,20 +1779,6 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 			p1 = p2;
 			len1 = strlen(p2) + 1;
 		}
-		/*
-		 * If PREPARE or extended query protocol commands caused error,
-		 * remove the temporary saved message.
-		 */
-		else
-		{
-			if (session_context->uncompleted_message)
-			{
-				pool_add_sent_message(session_context->uncompleted_message);
-				pool_remove_sent_message(session_context->uncompleted_message->kind,
-										 session_context->uncompleted_message->name);
-				session_context->uncompleted_message = NULL;
-			}
-		}
 
 		pool_write(frontend, "C", 1);
 		sendlen = htonl(len1+4);
@@ -1843,115 +1797,11 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
-								 POOL_CONNECTION_POOL *backend)
-{
-	int len, len1 = 0;
-	char *p = NULL;
-	char *p1 = NULL;
-	int status;
-	int sendlen;
-	int i;
-
-	POOL_SESSION_CONTEXT *session_context;
-	int num_params, send_num_params, num_dmy;
-	char kind = 't';
-
-	session_context = pool_get_session_context();
-	if (!session_context)
-	{
-		pool_error("ParameterDescription: cannot get session context");
-		return POOL_END;
-	}
-
-	/* only in replication mode and rewritten query */
-	if (!REPLICATION || !session_context->query_context->rewritten_query)
-		return SimpleForwardToFrontend('t', frontend, backend);
-
-	/* get number of parameters in original query */
-	num_params = session_context->query_context->num_original_params;
-
-	status = pool_read(MASTER(backend), &len, sizeof(len));
-	if (status < 0)
-	{
-		pool_error("ParameterDescription: error while reading message length");
-		return POOL_END;
-	}
-
-	len = ntohl(len);
-	len -= sizeof(int32);
-	len1 = len;
-
-	/* number of parameters in rewritten query is just discarded */
-	status = pool_read(MASTER(backend), &num_dmy, sizeof(int16));
-	len -= sizeof(int16);
-
-	p = pool_read2(MASTER(backend), len);
-	if (p == NULL)
-		return POOL_END;
-
-	p1 = malloc(len);
-	if (p1 == NULL)
-	{
-		pool_error("ParameterDescription: malloc failed");
-		return POOL_ERROR;
-	}
-	memcpy(p1, p, len);
-
-	for (i=0;i<NUM_BACKENDS;i++)
-	{
-		if (VALID_BACKEND(i) && !IS_MASTER_NODE_ID(i))
-		{
-			status = pool_read(CONNECTION(backend, i), &len, sizeof(len));
-			if (status < 0)
-			{
-				pool_error("ParameterDescription: error while reading message length");
-                free(p1);
-				return POOL_END;
-			}
-
-			len = ntohl(len);
-			len -= sizeof(int32);
-
-			p = pool_read2(CONNECTION(backend, i), len);
-			if (p == NULL)
-            {
-                free(p1);
-				return POOL_END;
-            }
-			if (len != len1)
-			{
-				pool_debug("ParameterDescription: length does not match between backends master(%d) %d th backend(%d) kind:(%c)",
-						   len, i, len1, kind);
-			}
-		}
-	}
-
-	pool_write(frontend, &kind, 1);
-
-	/* send back OIDs of parameters in original query and left are discarded */
-	len = sizeof(int16) + num_params * sizeof(int32);
-	sendlen = htonl(len + sizeof(int32));
-	pool_write(frontend, &sendlen, sizeof(int32));
-
-	send_num_params = htons(num_params);
-	pool_write(frontend, &send_num_params, sizeof(int16));
-
-	if (pool_write_and_flush(frontend, p1, num_params * sizeof(int32)) < 0)
-	{
-		pool_error("ParameterDescription: pool_write_and_flush failed");
-		free(p1);
-		return POOL_END;
-	}
-
-	free(p1);
-	return POOL_CONTINUE;
-}
-
 POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS ret;
+	POOL_SESSION_CONTEXT *session_context;
 
 	ret = SimpleForwardToFrontend('E', frontend, backend);
 	if (ret != POOL_CONTINUE)
@@ -2016,6 +1866,21 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 	}
 #endif
 
+	/* An error occurred with PREPARE or DEALLOCATE command.
+	 * Free pending portal object.
+	 */
+	session_context = pool_get_session_context();
+	if (session_context)
+	{
+		if (session_context->uncompleted_message)
+		{
+			pool_add_sent_message(session_context->uncompleted_message);
+			pool_remove_sent_message(session_context->uncompleted_message->kind,
+									 session_context->uncompleted_message->name);
+			session_context->uncompleted_message = NULL;
+		}
+	}
+
 	return POOL_CONTINUE;
 }
 
@@ -2129,7 +1994,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	session_context = pool_get_session_context();
 	if (!session_context)
 	{
-		pool_error("ProcessFrontendResponse: cannot get session context");
+		pool_error("Parse: cannot get session context");
 		return POOL_END;
 	}
 
@@ -2139,7 +2004,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
 		pool_log("ProcessFrontendResponse: failed to read kind from frontend. frontend abnormally exited");
-		return POOL_END_WITH_FRONTEND_ERROR;
+		return POOL_END;
 	}
 
 	pool_debug("ProcessFrontendResponse: kind from frontend %c(%02x)", fkind, fkind);
@@ -2162,15 +2027,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		return POOL_END;
 
 	if (fkind != 'S' && pool_is_ignore_till_sync())
-	{
-		/*
-		 * Flag setting for calling ProcessBackendResponse()
-		 * in pool_process_query().
-		 */
-		if (!pool_is_query_in_progress())
-			pool_set_query_in_progress();
 		return POOL_CONTINUE;
-	}
 
 	pool_unset_doing_extended_query_message();
 
@@ -2180,7 +2037,6 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		char *query;
 		Node *node;
 		List *parse_tree_list;
-		POOL_MEMORY_POOL *old_context = pool_memory;
 
 		case 'X':	/* Terminate */
 			return POOL_END;
@@ -2246,10 +2102,9 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 				return POOL_END;
 			}
 			query = "INSERT INTO foo VALUES(1)";
-			pool_memory = query_context->memory_context;
 			parse_tree_list = raw_parser(query);
 			node = (Node *) lfirst(list_head(parse_tree_list));
-			pool_start_query(query_context, query, node);
+			pool_start_query(query_context, query, strlen(query) + 1, node);
 			pool_where_to_send(query_context, query_context->original_query,
 							   query_context->parse_tree);
 
@@ -2257,8 +2112,6 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 				status = FunctionCall3(frontend, backend, len, contents);
 			else
 				status = FunctionCall(frontend, backend);
-
-			pool_memory = old_context;
 			break;
 
 		case 'c':	/* CopyDone */
@@ -2299,8 +2152,6 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 	if (pool_is_ignore_till_sync())
 	{
-		if (pool_is_query_in_progress())
-			pool_unset_query_in_progress();
 		return POOL_CONTINUE;
 	}
 
@@ -2361,7 +2212,7 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 			case 'E':	/* ErrorResponse */
 				status = ErrorResponse3(frontend, backend);
-				pool_set_command_success();
+				pool_unset_command_success();
 				if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID :
 						   REAL_MASTER_NODE_ID) != 'I')
 					pool_set_failed_transaction();
@@ -2379,10 +2230,6 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 					pool_unset_query_in_progress();
 				break;
 
-			case 't':	/* ParameterDescription */
-				status = ParameterDescription(frontend, backend);
-				break;
-
 			case 'I':	/* EmptyQueryResponse */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
 				if (pool_is_doing_extended_query_message())
@@ -2619,9 +2466,8 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 						id = pool_get_id(info, p1);
 						pool_debug("CopyDataRow: copying id: %d", id);
 						free(p1);
-						if (id < 0 || !VALID_BACKEND(id))
+						if (!VALID_BACKEND(id))
 						{
-							pool_error("CopyDataRow: pool_get_id returns invalid id: %d", id);
 							exit(1);
 						}
 						if (pool_write(CONNECTION(backend, id), &kind, 1))
@@ -2699,7 +2545,7 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 			return POOL_END;
 
 #ifdef DEBUG
-		strncpy(buf, string, len);
+		strlcpy(buf, string, sizeof(buf));
 		pool_debug("copy line %d %d bytes :%s:", j++, len, buf);
 #endif
 
@@ -3019,7 +2865,7 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 		if (qc->where_to_send[i] && statecmp(qc->query_state[i], POOL_PARSE_COMPLETE) < 0)
 		{
 			pool_debug("parse_before_bind: waiting for backend %d completing parse", i);
-			if (pool_send_and_wait(qc, contents, len, 1, i, "P") != POOL_CONTINUE)
+			if (pool_extended_send_and_wait(qc, "P", len, contents, 1, i) != POOL_CONTINUE)
 				return POOL_END;
 		}
 		else
